import unittest

from pcs_test.tools.color_text_runner.format import (
    Format,
    Output,
    separator1,
)
from pcs_test.tools.color_text_runner.writer import (
    DotWriter,
    ImprovedVerboseWriter,
    StandardVerboseWriter,
    TimeWriter,
    Writer,
)
from pcs_test.tools.parallel_test_runner import VanillaTextTestResult


class ColorTextTestResult(VanillaTextTestResult):
    slash_last = False
    traditional_verbose = False
    traceback_highlight = False
    fast_info = False
    rich_format = True
    measure_time = False

    def __init__(self, stream, descriptions, verbosity):
        super().__init__(stream, descriptions, verbosity)
        self.verbosity = 2 if self.traditional_verbose else verbosity

        self._format = Format(Output(self.rich_format))
        # pylint: disable=invalid-name
        self.reportWriter = self.__chooseWriter()(
            self.stream,
            self._format,
            self.descriptions,
            self.traceback_highlight,
            self.fast_info,
        )
        self.skip_map = {}

    def startTest(self, test):
        super(unittest.TextTestResult, self).startTest(test)
        self.reportWriter.startTest(test)

    def addSuccess(self, test):
        super(unittest.TextTestResult, self).addSuccess(test)
        self.reportWriter.addSuccess(test)

    def addError(self, test, err):
        super(unittest.TextTestResult, self).addError(test, err)
        self.reportWriter.addError(test, err, traceback=self.errors[-1][1])

    def addFailure(self, test, err):
        super(unittest.TextTestResult, self).addFailure(test, err)
        self.reportWriter.addFailure(test, err, traceback=self.failures[-1][1])

    def addSkip(self, test, reason):
        super(unittest.TextTestResult, self).addSkip(test, reason)
        self.skip_map.setdefault(reason, []).append(test)
        self.reportWriter.addSkip(test, reason)

    def addExpectedFailure(self, test, err):
        super(unittest.TextTestResult, self).addExpectedFailure(test, err)
        self.reportWriter.addExpectedFailure(test, err)

    def addUnexpectedSuccess(self, test):
        super(unittest.TextTestResult, self).addUnexpectedSuccess(test)
        self.reportWriter.addUnexpectedSuccess(test)

    def get_failed_names(self) -> list[str]:
        return [self._format.test_name(test) for test, _ in self.failures]

    def get_error_names(self) -> list[str]:
        return [self._format.test_name(test) for test, _ in self.errors]

    def get_errors(self) -> list[str]:
        return self._format.error_list(
            "ERROR", self.errors, self.descriptions, self.traceback_highlight
        )

    def get_failures(self) -> list[str]:
        return self._format.error_list(
            "FAIL", self.failures, self.descriptions, self.traceback_highlight
        )

    def get_skips(self) -> dict[str, int]:
        return {
            reason: len(test_list)
            for reason, test_list in self.skip_map.items()
        }

    def printErrors(self):
        error_lines = self.get_errors() + self.get_failures()
        if self.errors or self.failures or self.skip_map:
            error_lines.extend([separator1, ""])
        if self.errors + self.failures:
            error_lines.extend(
                [""]
                + self._format.error_overview(
                    self.errors, self.failures, self.slash_last
                )
            )

        if self.skip_map:
            error_lines.extend([""] + self._format.skip_overview(self.skip_map))

        if self.verbosity:
            error_lines.insert(0, "")

        for line in error_lines:
            self.stream.writeln(line)

    def __chooseWriter(self):  # pylint: disable=invalid-name
        if self.measure_time:
            return TimeWriter
        if self.traditional_verbose:
            return StandardVerboseWriter
        if self.verbosity > 1:
            return ImprovedVerboseWriter
        if self.verbosity > 0:
            return DotWriter
        return Writer


def get_text_test_result_class(
    *,
    slash_last_fail_in_overview=False,
    traditional_verbose=False,
    traceback_highlight=False,
    fast_info=False,
    rich_format=True,
    measure_time=False,
):
    ColorTextTestResult.slash_last = slash_last_fail_in_overview
    ColorTextTestResult.traditional_verbose = traditional_verbose
    ColorTextTestResult.traceback_highlight = traceback_highlight
    ColorTextTestResult.fast_info = fast_info
    ColorTextTestResult.rich_format = rich_format
    ColorTextTestResult.measure_time = measure_time

    return ColorTextTestResult



from unittest import TestCase

from pcs import settings
from pcs.common import reports
from pcs.common.auth import HostAuthData, HostWithTokenAuthData
from pcs.common.host import Destination
from pcs.lib.auth import validations

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal


class ValidateHostsWithToken(TestCase):
    def test_valid_no_reports(self):
        report_list = validations.validate_hosts_with_token(
            {
                "node1": HostWithTokenAuthData(
                    "TOKEN", [Destination("node1", 2224)]
                ),
                "node2": HostWithTokenAuthData(
                    "TOKEN", [Destination("node2", 2224)]
                ),
            }
        )
        assert_report_item_list_equal(report_list, [])

    def test_invalid_name(self):
        report_list = validations.validate_hosts_with_token(
            {
                "node1": HostWithTokenAuthData(
                    "TOKEN", [Destination("node1", 2224)]
                ),
                "": HostWithTokenAuthData(
                    "TOKEN", [Destination("node2", 2224)]
                ),
            }
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="host name",
                    option_value="",
                    allowed_values="",
                    cannot_be_empty=True,
                    forbidden_characters=None,
                )
            ],
        )

    def test_empty_token(self):
        report_list = validations.validate_hosts_with_token(
            {
                "node1": HostWithTokenAuthData(
                    "TOKEN", [Destination("node1", 2224)]
                ),
                "node2": HostWithTokenAuthData(
                    "", [Destination("node2", 2224)]
                ),
            }
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="token",
                    option_value="",
                    allowed_values="a string (min length: 1) (max length: 512)",
                    cannot_be_empty=True,
                    forbidden_characters=None,
                )
            ],
        )

    def test_token_too_long(self):
        bad_token = "a" * (settings.pcsd_token_max_chars + 1)
        report_list = validations.validate_hosts_with_token(
            {
                "node1": HostWithTokenAuthData(
                    "TOKEN", [Destination("node1", 2224)]
                ),
                "node2": HostWithTokenAuthData(
                    bad_token, [Destination("node2", 2224)]
                ),
            }
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="token",
                    option_value=bad_token,
                    allowed_values="a string (min length: 1) (max length: 512)",
                    cannot_be_empty=True,
                    forbidden_characters=None,
                )
            ],
        )

    def test_empty_dest_list(self):
        report_list = validations.validate_hosts_with_token(
            {
                "node1": HostWithTokenAuthData(
                    "TOKEN", [Destination("node1", 2224)]
                ),
                "node2": HostWithTokenAuthData("TOKEN", []),
            }
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="dest_list",
                    option_value="[]",
                    allowed_values="non-empty list of destinations for node 'node2'",
                    cannot_be_empty=True,
                    forbidden_characters=None,
                )
            ],
        )

    def test_empty_node_addr(self):
        report_list = validations.validate_hosts_with_token(
            {
                "node1": HostWithTokenAuthData(
                    "TOKEN", [Destination("node1", 2224), Destination("", 2224)]
                ),
            }
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="addr",
                    option_value="",
                    allowed_values="address for node 'node1'",
                    cannot_be_empty=True,
                    forbidden_characters=None,
                )
            ],
        )

    def test_invalid_port_number(self):
        report_list = validations.validate_hosts_with_token(
            {
                "node1": HostWithTokenAuthData(
                    "TOKEN",
                    [
                        Destination("node1", 2224),
                        Destination("a", -1),
                        Destination("b", 0),
                        Destination("c", 1000000000),
                    ],
                ),
            }
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="port for node 'node1'",
                    option_value=-1,
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="port for node 'node1'",
                    option_value=0,
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="port for node 'node1'",
                    option_value=1000000000,
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )


class ValidateHosts(TestCase):
    def test_valid_no_reports(self):
        report_list = validations.validate_hosts(
            {
                "node1": HostAuthData(
                    "username", "password", [Destination("node1", 2224)]
                ),
                "node2": HostAuthData(
                    "username", "password", [Destination("node2", 2224)]
                ),
            }
        )
        assert_report_item_list_equal(report_list, [])

    def test_invalid_name(self):
        report_list = validations.validate_hosts(
            {
                "node1": HostAuthData(
                    "username", "password", [Destination("node1", 2224)]
                ),
                "": HostAuthData(
                    "username", "password", [Destination("node2", 2224)]
                ),
            }
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="host name",
                    option_value="",
                    allowed_values="",
                    cannot_be_empty=True,
                    forbidden_characters=None,
                )
            ],
        )

    def test_empty_dest_list(self):
        report_list = validations.validate_hosts(
            {
                "node1": HostAuthData(
                    "username", "password", [Destination("node1", 2224)]
                ),
                "node2": HostAuthData("username", "password", []),
            }
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="dest_list",
                    option_value="[]",
                    allowed_values="non-empty list of destinations for node 'node2'",
                    cannot_be_empty=True,
                    forbidden_characters=None,
                )
            ],
        )

    def test_empty_node_addr(self):
        report_list = validations.validate_hosts(
            {
                "node1": HostAuthData(
                    "username",
                    "password",
                    [Destination("node1", 2224), Destination("", 2224)],
                ),
            }
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="addr",
                    option_value="",
                    allowed_values="address for node 'node1'",
                    cannot_be_empty=True,
                    forbidden_characters=None,
                )
            ],
        )

    def test_invalid_port_number(self):
        report_list = validations.validate_hosts_with_token(
            {
                "node1": HostAuthData(
                    "username",
                    "password",
                    [
                        Destination("node1", 2224),
                        Destination("a", -1),
                        Destination("b", 0),
                        Destination("c", 1000000000),
                    ],
                ),
            }
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="port for node 'node1'",
                    option_value=-1,
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="port for node 'node1'",
                    option_value=0,
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="port for node 'node1'",
                    option_value=1000000000,
                    allowed_values="a port number (1..65535)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )



# ruff: noqa: PLC0415 `import` should be at the top-level of a file
import sys
import unittest
from dataclasses import dataclass, field
from io import StringIO


class ParallelTestManager:
    def __init__(self, result_class, verbosity: int):
        self.result_class = result_class
        self.verbosity = verbosity

    def run_test(self, test_name: str) -> "ParallelTestResult":
        test = unittest.defaultTestLoader.loadTestsFromName(test_name)
        test_runner = ParallelTestRunner(
            verbosity=self.verbosity, resultclass=self.result_class
        )
        result = test_runner.run(test)
        return ParallelTestResult(
            result.testsRun,
            result.wasSuccessful(),
            result.get_errors(),
            result.get_failures(),
            result.get_skips(),
            result.get_error_names(),
            result.get_failed_names(),
            len(result.errors),
            len(result.failures),
            len(result.skipped),
            len(result.expectedFailures),
            len(result.unexpectedSuccesses),
        )


class VanillaTextTestResult(unittest.TextTestResult):
    # pylint: disable=no-self-use
    def get_failed_names(self) -> list[str]:
        return []

    def get_error_names(self) -> list[str]:
        return []

    def get_errors(self) -> list[str]:
        return self._get_fail_descriptions("ERROR", self.errors)

    def get_failures(self) -> list[str]:
        return self._get_fail_descriptions("FAIL", self.failures)

    def get_skips(self) -> dict[str, int]:
        return {}

    def _get_fail_descriptions(
        self, severity: str, item_list: list[tuple[unittest.TestCase, str]]
    ) -> list[str]:
        line_list = []
        for test, err in item_list:
            line_list.append(self.separator1)
            line_list.append(f"{severity}: {self.getDescription(test)}")
            line_list.append(self.separator2)
            line_list.append(err)
        return line_list

    def stopTest(self, test):
        del test
        # super().stopTest(test)
        if self.stream.seekable():
            self.stream.seek(0)
        if self.stream.readable():
            sys.stderr.write(self.stream.read())
        sys.stderr.flush()


@dataclass
class ParallelTestResult:
    """Representation of test result that can be pickled"""

    # pylint: disable=too-many-instance-attributes
    tests_run: int = 0
    was_successful: bool = True
    error_reports: list[str] = field(default_factory=list)
    failure_reports: list[str] = field(default_factory=list)
    skip_reports: dict[str, int] = field(default_factory=dict)
    error_names: list[str] = field(default_factory=list)
    failure_names: list[str] = field(default_factory=list)
    error_count: int = 0
    failure_count: int = 0
    skip_count: int = 0
    expected_failure_count: int = 0
    unexpected_success_count: int = 0

    def print_summary(
        self,
        time: float,
        vanilla: bool,
        last_slash: bool,
    ) -> None:
        # pylint: disable=import-outside-toplevel too-many-branches
        summary_lines = []
        if self.error_count or self.failure_count or self.skip_count or vanilla:
            summary_lines.append("")

        summary_lines.extend(self.error_reports)
        summary_lines.extend(self.failure_reports)

        if not vanilla:
            from pcs_test.tools.color_text_runner.format import Output

            output = Output(True)
            if self.error_count or self.failure_count or self.skip_count:
                summary_lines.append(unittest.TextTestResult.separator1)
                summary_lines.append("")
            if self.error_count or self.failure_count:
                summary_lines.append("")
                summary_lines.append(
                    output.red(
                        "for running failed tests only (errors are first then "
                        "failures):"
                    )
                )
                error_overview = " \\\n".join(
                    output.lightgrey(err)
                    for err in self.error_names + self.failure_names
                )
                if last_slash:
                    error_overview = f"{error_overview} \\"
                summary_lines.append("")
                summary_lines.append(error_overview)
            if self.skip_count:
                summary_lines.append("")
                summary_lines.append(
                    output.blue("Some tests have been skipped:")
                )
                summary_lines.append(
                    "\n".join(
                        [
                            f"{reason}\n  ({count}x)"
                            for reason, count in sorted(
                                self.skip_reports.items()
                            )
                        ]
                    )
                )
            summary_lines.append("")

        summary_lines.append(unittest.TextTestResult.separator2)
        summary_lines.append(
            f"Ran {self.tests_run} test{'s' if self.tests_run != 1 else ''} in "
            f"{time:.3f}s"
        )
        summary_lines.append("")

        infos = []
        if self.failure_count:
            infos.append(f"failures={self.failure_count}")
        if self.error_count:
            infos.append(f"errors={self.error_count}")
        if self.skip_count:
            infos.append(f"skipped={self.skip_count}")
        if self.expected_failure_count:
            infos.append(f"expected failures={self.expected_failure_count}")
        if self.unexpected_success_count:
            infos.append(
                f"unexpected successes={self.unexpected_success_count}"
            )
        info_line = "{result}{infos}".format(
            result="OK" if self.was_successful else "FAILED",
            infos=f" ({', '.join(infos)})" if infos else "",
        )
        summary_lines.append(info_line)
        print("\n".join(summary_lines), file=sys.stderr)


class ParallelTestRunner(unittest.TextTestRunner):
    def _makeResult(self):
        return self.resultclass(
            # pylint: disable=protected-access
            unittest.runner._WritelnDecorator(StringIO()),
            self.descriptions,
            self.verbosity,
        )

    def run(self, test):
        result = self._makeResult()
        result.failfast = self.failfast
        result.buffer = self.buffer
        start_test_run = getattr(result, "startTestRun", None)
        if start_test_run is not None:
            start_test_run()
        try:
            test(result)
        finally:
            stop_test_run = getattr(result, "stopTestRun", None)
            if stop_test_run is not None:
                stop_test_run()
        return result


def aggregate_test_results(
    results: list[ParallelTestResult],
) -> ParallelTestResult:
    result = ParallelTestResult()
    for res in results:
        result.tests_run += res.tests_run
        result.error_reports.extend(res.error_reports)
        result.failure_reports.extend(res.failure_reports)
        result.error_names.extend(res.error_names)
        result.failure_names.extend(res.failure_names)
        result.was_successful = result.was_successful and res.was_successful
        result.error_count += res.error_count
        result.failure_count += res.failure_count
        result.skip_count += res.skip_count
        result.expected_failure_count += res.expected_failure_count
        result.unexpected_success_count += res.unexpected_success_count
        for reason, count in res.skip_reports.items():
            result.skip_reports[reason] = (
                result.skip_reports.get(reason, 0) + count
            )
    result.error_names = sorted(set(result.error_names))
    result.failure_names = sorted(set(result.failure_names))
    return result



import shlex
from collections import defaultdict
from typing import (
    Callable,
    Iterable,
)

from pcs.cli.common.output import (
    INDENT_STEP,
    pairs_to_cmd,
)
from pcs.cli.reports.output import warn
from pcs.cli.rule import rule_expression_dto_to_lines
from pcs.common.pacemaker.constraint import (
    CibConstraintLocationAttributesDto,
    CibConstraintLocationDto,
    CibConstraintLocationSetDto,
)
from pcs.common.pacemaker.rule import CibRuleExpressionDto
from pcs.common.pacemaker.tools import (
    abs_score,
    is_negative_score,
)
from pcs.common.str_tools import (
    format_optional,
    indent,
    pairs_to_text,
)

from . import set as _set


def _attributes_to_pairs(
    attributes_dto: CibConstraintLocationAttributesDto,
) -> list[tuple[str, str]]:
    pairs = []
    if attributes_dto.resource_discovery:
        pairs.append(
            ("resource-discovery", str(attributes_dto.resource_discovery))
        )
    return pairs


def _attributes_to_text(
    attributes_dto: CibConstraintLocationAttributesDto,
    with_id: bool,
) -> list[str]:
    result = pairs_to_text(_attributes_to_pairs(attributes_dto))
    if attributes_dto.rules:
        result.append("Rules:")
        for rule_dto in attributes_dto.rules:
            result.extend(
                indent(
                    rule_expression_dto_to_lines(rule_dto, with_ids=with_id),
                    indent_step=INDENT_STEP,
                )
            )

    if attributes_dto.lifetime:
        result.append("Lifetime:")
        for rule_dto in attributes_dto.lifetime:
            result.extend(
                indent(
                    rule_expression_dto_to_lines(rule_dto, with_ids=with_id),
                    indent_step=INDENT_STEP,
                )
            )
    return result


def plain_constraint_to_text(
    constraint_dto: CibConstraintLocationDto,
    with_id: bool,
) -> list[str]:
    prefers_part = ""
    score = constraint_dto.attributes.score or "INFINITY"
    if not constraint_dto.attributes.rules:
        prefers_part = " {prefers} node '{node}' with score {score}".format(
            prefers=("avoids" if is_negative_score(score) else "prefers"),
            node=constraint_dto.attributes.node,
            score=abs_score(score),
        )

    result = [
        "{role}resource{resource_pattern} '{resource}'{prefers_part}".format(
            role=format_optional(constraint_dto.role),
            resource_pattern=("" if constraint_dto.resource_id else " pattern"),
            resource=(
                constraint_dto.resource_id or constraint_dto.resource_pattern
            ),
            prefers_part=prefers_part,
        )
    ]
    if with_id:
        result[0] += f" (id: {constraint_dto.attributes.constraint_id})"
    result.extend(
        indent(
            _attributes_to_text(constraint_dto.attributes, with_id),
            indent_step=INDENT_STEP,
        )
    )

    return result


def set_constraint_to_text(
    constraint_dto: CibConstraintLocationSetDto,
    with_id: bool,
) -> list[str]:
    return _set.set_constraint_to_text(
        constraint_dto.attributes.constraint_id,
        _attributes_to_text(constraint_dto.attributes, with_id),
        constraint_dto.resource_sets,
        with_id,
    )


def constraints_to_text(
    plain_dtos: Iterable[CibConstraintLocationDto],
    set_dtos: Iterable[CibConstraintLocationSetDto],
    with_id: bool,
) -> list[str]:
    result = []
    if plain_dtos:
        result.append("Location Constraints:")
        for constraint_dto in plain_dtos:
            result.extend(
                indent(
                    plain_constraint_to_text(constraint_dto, with_id),
                    indent_step=INDENT_STEP,
                )
            )
    if set_dtos:
        result.append("Location Set Constraints:")
        for set_constraint_dto in set_dtos:
            result.extend(
                indent(
                    set_constraint_to_text(set_constraint_dto, with_id),
                    indent_step=INDENT_STEP,
                )
            )
    return result


def _plain_constraint_get_resource_for_cmd(
    constraint_dto: CibConstraintLocationDto,
) -> str:
    if constraint_dto.resource_id:
        resource = f"resource%{constraint_dto.resource_id}"
    else:
        resource = f"regexp%{constraint_dto.resource_pattern}"
    return shlex.quote(resource)


def _plain_constraint_to_cmd(
    constraint_dto: CibConstraintLocationDto,
) -> list[str]:
    score = constraint_dto.attributes.score or "INFINITY"
    result = [
        "pcs -- constraint location add {id} {resource} {node} {score}".format(
            id=shlex.quote(constraint_dto.attributes.constraint_id),
            resource=_plain_constraint_get_resource_for_cmd(constraint_dto),
            node=shlex.quote(str(constraint_dto.attributes.node)),
            score=shlex.quote(f"score={score}"),
        )
    ]
    if constraint_dto.attributes.resource_discovery:
        result.extend(
            indent(
                [pairs_to_cmd(_attributes_to_pairs(constraint_dto.attributes))],
                indent_step=INDENT_STEP,
            )
        )
    return result


def _rule_to_cmd_pairs(rule: CibRuleExpressionDto) -> list[tuple[str, str]]:
    pairs = []
    if rule.options.get("role"):
        pairs.append(("role", rule.options["role"]))
    if rule.options.get("score"):
        pairs.append(("score", rule.options["score"]))
    elif rule.options.get("score-attribute"):
        pairs.append(("score-attribute", rule.options["score-attribute"]))
    return pairs


def _plain_constraint_rule_to_cmd(
    constraint_dto: CibConstraintLocationDto,
) -> list[list[str]]:
    if len(constraint_dto.attributes.rules) > 1:
        warn(
            f"Constraint '{constraint_dto.attributes.constraint_id}' contains "
            "more than one rule, which is no longer supported. Instead, each "
            "rule will be put in a separate constraint."
        )

    result = []
    for rule_index, one_rule in enumerate(constraint_dto.attributes.rules):
        command = [
            "pcs -- constraint location {resource} rule".format(
                resource=_plain_constraint_get_resource_for_cmd(constraint_dto)
            )
        ]
        pairs = [("id", one_rule.id)]
        if rule_index == 0:
            # Constraint ID must be unique. Since we split additional rules
            # into their own constraints, we cannot use the original constraint
            # ID for the newly created constraints.
            pairs.append(
                (
                    "constraint-id",
                    constraint_dto.attributes.constraint_id,
                )
            )
        command.extend(
            indent(
                [
                    pairs_to_cmd(
                        pairs
                        + _attributes_to_pairs(constraint_dto.attributes)
                        + _rule_to_cmd_pairs(one_rule)
                    ),
                    shlex.quote(one_rule.as_string),
                ],
                indent_step=INDENT_STEP,
            )
        )
        result.append(command)
    return result


def plain_constraint_to_cmd(
    constraint_dto: CibConstraintLocationDto,
) -> list[list[str]]:
    if constraint_dto.attributes.lifetime:
        warn(
            "Lifetime configuration detected in constraint "
            f"'{constraint_dto.attributes.constraint_id}' but not supported by "
            "this command."
            " Command for creating the constraint is omitted."
        )
        return []
    if constraint_dto.role:
        warn(
            f"Resource role detected in constraint "
            f"'{constraint_dto.attributes.constraint_id}' but not supported by "
            "this command."
            " Command for creating the constraint is omitted."
        )
        return []
    if constraint_dto.attributes.rules:
        return _plain_constraint_rule_to_cmd(constraint_dto)
    return [_plain_constraint_to_cmd(constraint_dto)]


def _split_based_on_score(
    constraint_dtos: Iterable[CibConstraintLocationDto],
) -> tuple[list[CibConstraintLocationDto], list[CibConstraintLocationDto]]:
    prefers: list[CibConstraintLocationDto] = []
    avoids: list[CibConstraintLocationDto] = []
    for constraint_dto in constraint_dtos:
        if constraint_dto.attributes.score is not None and is_negative_score(
            constraint_dto.attributes.score
        ):
            avoids.append(constraint_dto)
        else:
            prefers.append(constraint_dto)
    return prefers, avoids


def _per_resource_constraint_to_str(
    constraint_dto: CibConstraintLocationDto, with_ids: bool
) -> str:
    return "node '{node}' with score {score}{id}".format(
        node=constraint_dto.attributes.node,
        score=abs_score(constraint_dto.attributes.score or "0"),
        id=(
            f" (id: {constraint_dto.attributes.constraint_id})"
            if with_ids
            else ""
        ),
    )


def _split_with_rules(
    constraint_dtos: Iterable[CibConstraintLocationDto],
) -> tuple[list[CibConstraintLocationDto], list[CibConstraintLocationDto]]:
    without_rules: list[CibConstraintLocationDto] = []
    with_rules: list[CibConstraintLocationDto] = []
    for constraint_dto in constraint_dtos:
        if constraint_dto.attributes.rules:
            with_rules.append(constraint_dto)
        else:
            without_rules.append(constraint_dto)
    return without_rules, with_rules


def _split_patterns(
    constraint_dtos: Iterable[CibConstraintLocationDto],
) -> tuple[list[CibConstraintLocationDto], list[CibConstraintLocationDto]]:
    ids: list[CibConstraintLocationDto] = []
    patterns: list[CibConstraintLocationDto] = []
    for constraint_dto in constraint_dtos:
        if constraint_dto.resource_pattern:
            patterns.append(constraint_dto)
        else:
            ids.append(constraint_dto)
    return ids, patterns


def _labeled_constraints_list_to_text(
    label: str,
    constraint_dtos: Iterable[CibConstraintLocationDto],
    render_fn: Callable[[CibConstraintLocationDto], str],
) -> list[str]:
    result = []
    if constraint_dtos:
        result.extend(
            [f"{label}:"]
            + indent(
                [
                    render_fn(constraint_dto)
                    for constraint_dto in constraint_dtos
                ],
                indent_step=INDENT_STEP,
            ),
        )
    return result


def _with_rules_to_text(
    constraint: CibConstraintLocationDto, with_ids: bool
) -> list[str]:
    result = [
        "Constraint:{id}".format(
            id=f" {constraint.attributes.constraint_id}" if with_ids else ""
        )
    ]
    rules_lines = []
    for rule_dto in constraint.attributes.rules:
        rules_lines.extend(
            indent(
                rule_expression_dto_to_lines(rule_dto, with_ids),
                indent_step=INDENT_STEP,
            )
        )
    result.extend(indent(["Rules:"] + rules_lines, indent_step=INDENT_STEP))
    return result


def _per_resource_constraints_to_text(
    constraint_dtos: Iterable[CibConstraintLocationDto], with_ids: bool
) -> list[str]:
    def render_fn(constraint: CibConstraintLocationDto) -> str:
        return _per_resource_constraint_to_str(constraint, with_ids)

    result = []
    without_rules, with_rules = _split_with_rules(constraint_dtos)
    prefers, avoids = _split_based_on_score(without_rules)
    result.extend(
        _labeled_constraints_list_to_text("Prefers", prefers, render_fn)
    )
    result.extend(
        _labeled_constraints_list_to_text("Avoids", avoids, render_fn)
    )
    for constraint in with_rules:
        result.extend(_with_rules_to_text(constraint, with_ids))
    return result


def constraints_to_grouped_by_resource_text(
    constraint_dtos: Iterable[CibConstraintLocationDto],
    with_ids: bool,
) -> list[str]:
    result = []
    patterns: dict[str, list[CibConstraintLocationDto]] = defaultdict(list)
    resources: dict[str, list[CibConstraintLocationDto]] = defaultdict(list)
    for constraint_dto in constraint_dtos:
        if constraint_dto.resource_pattern:
            patterns[constraint_dto.resource_pattern].append(constraint_dto)
        else:
            resources[str(constraint_dto.resource_id)].append(constraint_dto)
    for resource_id, constraints in resources.items():
        result.append(f"Resource: {resource_id}")
        result.extend(
            indent(
                _per_resource_constraints_to_text(constraints, with_ids),
                indent_step=INDENT_STEP,
            )
        )
    for resource_pattern, constraints in patterns.items():
        result.append(f"Resource pattern: {resource_pattern}")
        result.extend(
            indent(
                _per_resource_constraints_to_text(constraints, with_ids),
                indent_step=INDENT_STEP,
            )
        )

    return result


def _per_node_constraint_to_str(
    constraint_dto: CibConstraintLocationDto, with_ids: bool
) -> str:
    return "resource{pattern} '{id}' with score {score}{constraint_id}".format(
        pattern=" pattern" if constraint_dto.resource_pattern else "",
        id=constraint_dto.resource_pattern or constraint_dto.resource_id,
        score=abs_score(constraint_dto.attributes.score or "0"),
        constraint_id=(
            f" (id: {constraint_dto.attributes.constraint_id})"
            if with_ids
            else ""
        ),
    )


def constraints_to_grouped_by_node_text(
    constraints_dtos: Iterable[CibConstraintLocationDto],
    with_ids: bool,
) -> list[str]:
    constraint_dtos, with_rules = _split_with_rules(constraints_dtos)
    if with_rules:
        warn("Constraints with rules are not displayed.")
    nodes: dict[str, list[CibConstraintLocationDto]] = defaultdict(list)
    for constraint in constraint_dtos:
        if constraint.attributes.node:
            nodes[constraint.attributes.node].append(constraint)
    result = []

    def _render_fn(constraint: CibConstraintLocationDto) -> str:
        return _per_node_constraint_to_str(constraint, with_ids)

    for node, constraints in nodes.items():
        result.extend([f"Node: {node}"])
        prefers, avoids = _split_based_on_score(constraints)
        result.extend(
            indent(
                _labeled_constraints_list_to_text(
                    "Preferred by", prefers, _render_fn
                ),
                indent_step=INDENT_STEP,
            )
        )
        result.extend(
            indent(
                _labeled_constraints_list_to_text(
                    "Avoided by", avoids, _render_fn
                ),
                indent_step=INDENT_STEP,
            )
        )
    return result
